{****************************************************}
{* Common routines unit for LogicGates7 application *}
{****************************************************}

unit common;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, ExtCtrls, Dialogs, Graphics;

type
  TBooleans = array of Boolean;
  TSwitchData = record                                                         // switches data (cf. in multipliers units for details)
    Name: string;
    Data: array[0..2] of Integer;
  end;
  TSwitches = array of record                                                  // switches objects (transparent image and bar shapes)
    Switch0: TImage;
    Switch1, Switch2: TShape;
  end;
  TLeds = array of TShape;                                                     // LED objects (shape)
  TStatus = array of Boolean;                                                  // switch/LED status (on/off)

const
  // UTF8 values for digits subscripts
  SUB_Digits: array[0..9] of string = (
    #$E2#$82#$80, #$E2#$82#$81, #$E2#$82#$82, #$E2#$82#$83, #$E2#$82#$84, #$E2#$82#$85, #$E2#$82#$86, #$E2#$82#$87, #$E2#$82#$88, #$E2#$82#$89
  );

// --------------------------------- Public procedures (switches, LEDs, multipliers) --------------------------------
procedure SwitchOnOff(DSwitches: array of TSwitchData; Switch0: TImage; Switch1, Switch2: TShape; SwitchOn: Boolean);
procedure SwitchesOnOff(DSwitches: array of TSwitchData; Switches: TSwitches; Status: TStatus);
procedure LedOnOff(Led: TShape; LedOn: Boolean);
procedure LedsOnOff(Leds: TLeds; Status: TStatus);
procedure Multiplier2x2(A0, A1, B0, B1: Boolean; out C0, C1, C2, C3: Boolean);
procedure Multiplier4x2(var InputA, InputB: array of Boolean; out Output: TBooleans);
procedure Multiplier4x3(var InputA, InputB: array of Boolean; out Output: TBooleans);
// ------------------------------------------------------------------------------------------------------------------

implementation

{ Set given switch to position "on" or "off" }

procedure SwitchOnOff(DSwitches: array of TSwitchData; Switch0: TImage; Switch1, Switch2: TShape; SwitchOn: Boolean);

var
  I, IX: Integer;

begin
  // Determine index for switch data array by looking up the actual switch image object's name
  for I := 0 to Length(DSwitches) - 1 do begin
    if DSwitches[I].Name = Copy(Switch0.Name, 3, Length(Switch0.Name)) then
      IX := I;
  end;
  // Switch's vertical "off" position (defined in its data)
  Switch0.Top := DSwitches[IX].Data[0]; Switch1.Top := DSwitches[IX].Data[1]; Switch2.Top := DSwitches[IX].Data[2];
  if SwitchOn then begin
    // Adjust switch object's vertical position if the switch has to be set "on"
    Switch0.Top := Switch0.Top + 10;
    Switch1.Top := Switch1.Top + 10;
    Switch2.Top := Switch2.Top + 10;
  end;
end;

{ Set all switches to "on" or "off" position (depending on status values) }

procedure SwitchesOnOff(DSwitches: array of TSwitchData; Switches: TSwitches; Status: TStatus);

var
  I: Integer;

begin
  for I := 0 to Length(Switches) - 1 do
    // Set position for each switch (array element)
    SwitchOnOff(DSwitches, Switches[I].Switch0, Switches[I].Switch1, Switches[I].Switch2, Status[I]);
end;

{ Turn a given LED "on" or "off" }

procedure LedOnOff(Led: TShape; LedOn: Boolean);

begin
  if LedOn then
    // LED on shown by solid color
    Led.Brush.Style := bsSolid
  else
    // LED off shown by pattern
    Led.Brush.Style := bsDiagCross;
end;

{ Turn all LEDs "on" or "off" (depending on status values) }

procedure LedsOnOff(Leds: TLeds; Status: TStatus);

var
  I: Integer;

begin
  for I := 0 to Length(Leds) - 1 do begin
    // Turn each LED (array element) "on" or "off"
    LedOnOff(Leds[I], Status[I]);
  end;
end;

{ 1-bit-half-adder }

procedure HalfAdder(A, B: Boolean; out S, C: Boolean);

// Calculation done as does the real circuit: Usage of an XOR and an AND gate

begin
  S := A xor B;
  C := A and B;
end;

{ 1-bit-full-adder }

procedure Adder(A, B, C: Boolean; out S1, C1, S2, C2, Sout, Cout: Boolean);

// Calculation done as does the real circuit: Usage of two half-adders and an Or gate

begin
  HalfAdder(A, B, S1, C1);
  HalfAdder(C, S1, S2, C2);
  Sout := S2; Cout := C1 or C2;
end;

{ 4-bits-full-adder }

procedure FullAdder4(var Input1, Input2: array of Boolean; out Sum: array of Boolean; out Carry: Boolean);

var
  I: Integer;
  A, B, C, S1, S2, C1, C2, Sout, Cout: Boolean;

begin
  // Do the addition, using 4 1-bit full-adders, as is done in a real circuit:
  // The sum output of a given adder will the sum value for that bit,
  // the carry output being used as input for the subsequent adder
  Cout := False;                                                               // Cin for first adder set to 0V
  for I := 0 to 3 do begin
    A := Input1[I]; B := Input2[I];                                            // actual bit for A and B input
    C := Cout;                                                                 // Cin = previous Cout
    // 1-bit-full-adder binary addition
    Adder(A, B, C, S1, C1, S2, C2, Sout, Cout);
    // Actual bit sum LED status
    Sum[I] := SOut;
  end;
  Carry := Cout;
end;

{ 2x2-bits binary multiplier }

procedure Multiplier2x2(A0, A1, B0, B1: Boolean; out C0, C1, C2, C3: Boolean);

// Calculation done as does the real circuit: Usage of four AND-gates and two HA

var
  O00, O01, O10, O11, C: Boolean;

begin
  O00 := A0 and B0; O01 := A0 and B1;
  O10 := A1 and B0; O11 := A1 and B1;
  C0 := O00;
  HalfAdder(O01, O10, C1, C);
  HalfAdder(O11, C, C2, C3);
end;

{ 4x2-bits binary multiplier }

procedure Multiplier4x2(var InputA, InputB: array of Boolean; out Output: TBooleans);

// Calculation done as does the real circuit: Usage of 2 * 4 AND-gates and 2 - 1 adders

var
  I, J: Integer;
  GatesANDIn: array[0..1, 0..3, 0..1] of Boolean;
  GatesANDOut: array[0..1, 0..3] of Boolean;
  AdderInA, AdderInB: array[0..3] of Boolean;
  Sum: array[0..3] of Boolean;
  Carry: Boolean;

begin
  SetLength(Output, 6);
  for I := 0 to 1 do begin
    for J := 0 to 3 do begin
      GatesANDIn[I, J, 0] := InputA[I];                                        // A0, A1
      GatesANDIn[I, J, 1] := InputB[J];                                        // B0 - B3
      GatesANDOut[I, J] := GatesANDIn[I, J, 0] and GatesANDIn[I, J, 1];
    end;
  end;
  for J := 1 to 3 do
    AdderInA[J - 1] := GatesANDOut[0, J];                                      // Adder input A0 - A2
  AdderInA[3] := False;                                                        // Adder input A3 set to 0V
  for J := 0 to 3 do
    AdderInB[J] := GatesANDOut[1, J];                                          // Adder input B0 - B3
  FullAdder4(AdderInA, AdderInB, Sum, Carry);                                  // 4-bits-full-adder
  Output[0] := GatesANDOut[0, 0];                                              // C0 = A0 and B0
  for J := 1 to 4 do
    Output[J] := Sum[J - 1];                                                   // C1 - C4 = adder sum bits S0 - S3
  Output[5] := Carry;                                                          // C5 = adder carry
end;

{ 4x3-bits binary multiplier }

procedure Multiplier4x3(var InputA, InputB: array of Boolean; out Output: TBooleans);

// Calculation done as does the real circuit: Usage of 3 * 4 AND-gates and 3 - 1 adders

var
  I, J: Integer;
  GatesANDIn: array[0..2, 0..3, 0..1] of Boolean;
  GatesANDOut: array[0..2, 0..3] of Boolean;
  AdderInA, AdderInB: array[0..3] of Boolean;
  Sum: array[0..3] of Boolean;
  Carry: Boolean;

begin
  SetLength(Output, 7);
  for I := 0 to 2 do begin
    for J := 0 to 3 do begin
      GatesANDIn[I, J, 0] := InputA[I];                                        // A0, A1, A2
      GatesANDIn[I, J, 1] := InputB[J];                                        // B0 - B3
      GatesANDOut[I, J] := GatesANDIn[I, J, 0] and GatesANDIn[I, J, 1];
    end;
  end;
  for J := 1 to 3 do
    AdderInA[J - 1] := GatesANDOut[0, J];                                      // First adder input A0 - A2
  AdderInA[3] := False;                                                        // First adder input A3 set to 0V
  for J := 0 to 3 do
    AdderInB[J] := GatesANDOut[1, J];                                          // First adder input B0 - B3
  FullAdder4(AdderInA, AdderInB, Sum, Carry);                                  // 4-bits-full-adder
  Output[0] := GatesANDOut[0, 0];                                              // C0 = A0 and B0
  Output[1] := Sum[0];                                                         // C1 = first adder sum bit S0
  for J := 1 to 3 do
    AdderInA[J - 1] := Sum[J];                                                 // Second adder input A0 - A2 = first adder sum bits S1 - S3
  AdderInA[3] := Carry;                                                        // Second adder input A3 = first adder carry
  for J := 0 to 3 do
    AdderInB[J] := GatesANDOut[2, J];                                          // Second adder input for B (output of 3rd group of AND gates)
  FullAdder4(AdderInA, AdderInB, Sum, Carry);                                  // 4-bits-full-adder
  for J := 2 to 5 do
    Output[J] := Sum[J - 2];                                                   // C2 - C5 = second adder sum bits S0 - S3
  Output[6] := Carry;                                                          // C6 = second adder carry
end;

end.

